/*S2IBIS Version 1.0 - Steven Lipa, NC State University  August 4, 1994       */

/******************************************************************************/
/*  Copyright 1994, North Carolina State University.                          */
/*                                                                            */
/*  North Carolina State University (NCSU) and the author hereby disclaim     */
/*  all implied warranties, including the implied warranties of merchant-     */
/*  ability and fitness for a particular purpose. NCSU is not liable for      */
/*  any damages incurred by the use of this software and/or documentation,    */
/*  including direct, indirect, special, incidental, or consequential         */
/*  damages.                                                                  */
/*                                                                            */
/*  NCSU grants the bearer the right to modify, copy, and redistribute        */
/*  this software and documentation, but requests that NCSU be acknowledged   */
/*  for its contribution in the resulting product.                            */
/******************************************************************************/

#include "s2ibis.h"

extern PINPTR topofpinstack;
extern MODPTR topofmodstack;

char *strfind();

void printpinlist(PINPTR pinptr)
{

pinptr=topofpinstack;
fprintf(stderr,"************************************************************\n");
fprintf(stderr,"*                     s2ibis pin list                      *\n");
fprintf(stderr,"************************************************************\n");
while (pinptr->nextlink) 
    {
     fprintf(stderr,"pin %s\n",pinptr->nextlink->pin);
     fprintf(stderr,"spice_node %s\n",pinptr->nextlink->spice_node);
     fprintf(stderr,"signal_name %s\n",pinptr->nextlink->signal_name);
     fprintf(stderr,"model_name %s\n",pinptr->nextlink->model_name);
     fprintf(stderr,"R_pin %s\n",pinptr->nextlink->R_pin);
     fprintf(stderr,"L_pin %s\n",pinptr->nextlink->L_pin);
     fprintf(stderr,"C_pin %s\n",pinptr->nextlink->C_pin);
     fprintf(stderr,"model_type %d\n",pinptr->nextlink->model_type);
     fprintf(stderr,"polarity %d\n",pinptr->nextlink->polarity);
     fprintf(stderr,"show_polarity %d\n",pinptr->nextlink->show_polarity);
     fprintf(stderr,"polartext %s\n",pinptr->nextlink->polartext);
     fprintf(stderr,"inpin %s\n",pinptr->nextlink->inpin);
     fprintf(stderr,"vil %f\n",pinptr->nextlink->vil);
     fprintf(stderr,"vih %f\n",pinptr->nextlink->vih);
     fprintf(stderr,"vpwr_min %f\n",pinptr->nextlink->vpwr_min);
     fprintf(stderr,"vpwr_typ %f\n",pinptr->nextlink->vpwr_typ);
     fprintf(stderr,"vpwr_max %f\n",pinptr->nextlink->vpwr_max);
     fprintf(stderr,"enable %s\n",pinptr->nextlink->enable);
     fprintf(stderr,"show_enable %d\n",pinptr->nextlink->show_enable);
     fprintf(stderr,"enabletext %s\n",pinptr->nextlink->enabletext);
     fprintf(stderr,"forceto %f\n",pinptr->nextlink->forceto);
     fprintf(stderr,"ramp_sim_time %le\n",pinptr->nextlink->ramp_sim_time);
     fprintf(stderr,"c_comp %s",pinptr->nextlink->c_comp);
     fprintf(stderr,"Vinh %s\n",pinptr->nextlink->Vinh);
     fprintf(stderr,"Vinl %s\n",pinptr->nextlink->Vinl);
     fprintf(stderr,"vihrt %f\n",pinptr->nextlink->vihrt);
     fprintf(stderr,"vihrn %f\n",pinptr->nextlink->vihrn);
     fprintf(stderr,"vihrx %f\n",pinptr->nextlink->vihrx);
     fprintf(stderr,"vilrt %f\n",pinptr->nextlink->vilrt);
     fprintf(stderr,"vilrn %f\n",pinptr->nextlink->vilrn);
     fprintf(stderr,"vilrx %f\n",pinptr->nextlink->vilrx);
     fprintf(stderr,"trt %le\n",pinptr->nextlink->trt);
     fprintf(stderr,"trn %le\n",pinptr->nextlink->trn);
     fprintf(stderr,"trx %le\n",pinptr->nextlink->trx);
     fprintf(stderr,"tft %le\n",pinptr->nextlink->tft);
     fprintf(stderr,"tfn %le\n",pinptr->nextlink->tfn);
     fprintf(stderr,"tfx %le\n",pinptr->nextlink->tfx);
     fprintf(stderr,"no_model %d\n",pinptr->nextlink->no_model);
     fprintf(stderr,"************************************************************\n");
     pinptr=pinptr->nextlink;
    }
fprintf(stderr,"*                 end of s2ibis pin list                   *\n");
fprintf(stderr,"************************************************************\n");
}

void strncpyWarn( char * s1, char *s2, size_t n)
/* n is the size of the array */
{
   strncpy( s1,s2,n );

   if ( strlen( s2 ) >= n )
   {
      fprintf(stderr,"WARNING:  String \"%s\"'s expected length is : %d\n", s2, (n-1) ); 
      s1[n-1] = '\0';
   }
}
   
PINPTR addpin(char ps1[], char ps2[])
{
/* this routine adds a pin to the pin stack.  there is no limit to the size of the stack */
char pin_name[PIN_NAME_LENGTH];
char sig_name[SIGNAL_NAME_LENGTH];
char mod_name[MODEL_NAME_LENGTH],dummod[MODEL_NAME_LENGTH];
char r_pin[R_PIN_NAME_LENGTH];
char l_pin[L_PIN_NAME_LENGTH];
char c_pin[C_PIN_NAME_LENGTH];
char mod_type[MODEL_TYPE_LENGTH];
char spicenode[SPICE_NODE_NAME_LENGTH];
double rst;
float  vin_low;
float  vin_high;
float  vp_typ;
float  vp_min;
float  vp_max;
float  vgnd;
char  in_pin[PIN_NAME_LENGTH];
char  en_pin[PIN_NAME_LENGTH];
int  polar;
int  colsread;           /* used to help parse input strings (# of columns)*/

PINPTR pinptr;

/*find last pin on stack, allocate storage for new pin and put pointer to new pin*/
/*into the nextlink member of the last pin. */
pinptr=topofpinstack;
while (pinptr->nextlink) {
			 pinptr=pinptr->nextlink;
			}
pinptr->nextlink=(struct pinentry *) calloc(1,sizeof(PINELEM));
if (pinptr->nextlink == NULL)
   {
    fprintf(stderr,"s2ibis: Unable to allocate memory.  Exiting.\n");
    exit(-1);
   }
pinptr->nextlink->lastlink=pinptr;
/*set nextlink->nextlink to NULL *now* so if there is an error, at least the printpinlist  */
/*function will know when to stop!                                                         */
pinptr->nextlink->nextlink=NULL;

/*Parse the first line.  This is the same for all pins                                      */
colsread=sscanf(ps1,"%s %s %s %s %s %s",pin_name,sig_name,mod_name,r_pin,l_pin,c_pin);
strncpyWarn(pinptr->nextlink->pin,pin_name,sizeof( pinptr->nextlink->pin ));
strncpyWarn(pinptr->nextlink->signal_name,sig_name, sizeof( pinptr->nextlink->signal_name ) );
strncpyWarn(pinptr->nextlink->model_name,mod_name, sizeof( pinptr->nextlink->model_name ) );
if (colsread>3) strncpyWarn(pinptr->nextlink->R_pin,r_pin,sizeof( pinptr->nextlink->R_pin ) );
if (colsread>4) strncpyWarn(pinptr->nextlink->L_pin,l_pin,sizeof( pinptr->nextlink->L_pin ) );
if (colsread>5) strncpyWarn(pinptr->nextlink->C_pin,c_pin,sizeof( pinptr->nextlink->C_pin ) );

/* add the defaults for the various optional fields                                         */
pinptr->nextlink->show_polarity=1;
pinptr->nextlink->show_enable=1;
pinptr->nextlink->trt=0.1E-9;
pinptr->nextlink->trn=0.1E-9;
pinptr->nextlink->trx=0.1E-9;
pinptr->nextlink->tft=0.1E-9;
pinptr->nextlink->tfn=0.1E-9;
pinptr->nextlink->tfx=0.1E-9;
pinptr->nextlink->no_model=0;
strncpyWarn(pinptr->nextlink->c_comp,"           5.0pF         5.0pF         5.0pF\n", sizeof( pinptr->nextlink->c_comp ) ); 
strncpyWarn(pinptr->nextlink->inpin,"", sizeof( pinptr->nextlink->inpin ) ); /* not optional! just making sure it is null initially */ 
strncpyWarn(pinptr->nextlink->enable,"", sizeof( pinptr->nextlink->enable ) );/* not optional! just making sure it is null initially */
strncpyWarn(pinptr->nextlink->Vinh,"", sizeof( pinptr->nextlink->Vinh ) );
strncpyWarn(pinptr->nextlink->Vinl,"", sizeof( pinptr->nextlink->Vinl ) );
strncpyWarn(pinptr->nextlink->polartext,"", sizeof( pinptr->nextlink->polartext ) );
strncpyWarn(pinptr->nextlink->enabletext,"", sizeof( pinptr->nextlink->enabletext ) );

/*Parse the second line.  This varies depending on the pin type                             */
/* First read in the model type and figure out how to read the rest of the line             */
colsread=sscanf(ps2,"%s",mod_type);

/*Now, check out the model_type to see what to do next*/
/*model_type can be Input, Output, I/O, 3-state, Open_drain, POWER, GND, NC                 */

if ((mod_type[0]=='I') && (mod_type[1]=='n'))
  {
   pinptr->nextlink->model_type=1;  
   colsread=sscanf(ps2,"%s %s %d %f %f",dummod,spicenode,&polar,&vin_low,&vin_high);
   if (colsread != 5)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was Input.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode,sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->polarity=polar;
   pinptr->nextlink->vil=vin_low;
   pinptr->nextlink->vih=vin_high;
  }
else if ((mod_type[0]=='O') && (mod_type[1]=='u')) 
  {
   pinptr->nextlink->model_type=2;  
   colsread=sscanf(ps2,"%s %s %le %s",dummod,spicenode,&rst,in_pin);
   if (colsread != 4)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was Output.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode,sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->ramp_sim_time=rst;
   strncpyWarn(pinptr->nextlink->inpin,in_pin,sizeof( pinptr->nextlink->inpin ) );
  }
else if ((mod_type[0]=='I') && (mod_type[1]=='/'))
  {
   pinptr->nextlink->model_type=3;  
   colsread=sscanf(ps2,"%s %s %le %s %s %d %f %f",dummod,spicenode,&rst,in_pin,en_pin,&polar,&vin_low,&vin_high);
   if (colsread != 8)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was I/O.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode,sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->ramp_sim_time=rst;
   strncpyWarn(pinptr->nextlink->inpin,in_pin,sizeof( pinptr->nextlink->inpin ) );
   strncpyWarn(pinptr->nextlink->enable,en_pin, sizeof( pinptr->nextlink->enable ) );
   pinptr->nextlink->polarity=polar;
   pinptr->nextlink->vil=vin_low;
   pinptr->nextlink->vih=vin_high;
  }
else if (mod_type[0]=='3')
  {
   pinptr->nextlink->model_type=4;  
   colsread=sscanf(ps2,"%s %s %le %s %s",dummod,spicenode,&rst,in_pin,en_pin);
   if (colsread != 5)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was 3-state.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode, sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->ramp_sim_time=rst;
   strncpyWarn(pinptr->nextlink->inpin,in_pin, sizeof( pinptr->nextlink->inpin ));
   strncpyWarn(pinptr->nextlink->enable,en_pin, sizeof( pinptr->nextlink->enable ) );
  }
else if ((mod_type[0]=='O') && (mod_type[1]=='p'))
  {
   pinptr->nextlink->model_type=5;  
   colsread=sscanf(ps2,"%s %s %le %s",dummod,spicenode,&rst,in_pin);
   if (colsread != 4)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was Open_drain.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode, sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->ramp_sim_time=rst;
   strncpyWarn(pinptr->nextlink->inpin,in_pin, sizeof( pinptr->nextlink->inpin ) );
  }
else if (mod_type[0]=='P')
  {
   pinptr->nextlink->model_type=6;  
   colsread=sscanf(ps2,"%s %s %f %f %f",dummod,spicenode,&vp_typ,&vp_min,&vp_max);
   if (colsread < 3)
      {
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was POWER.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode, sizeof( pinptr->nextlink->spice_node ) );
   pinptr->nextlink->vpwr_typ=vp_typ;
   if (colsread>3) pinptr->nextlink->vpwr_min=vp_min;
   if (colsread>4) pinptr->nextlink->vpwr_max=vp_max;
  }
else if (mod_type[0]=='G')
  {
   pinptr->nextlink->model_type=7;  
   colsread=sscanf(ps2,"%s %s %f",dummod,spicenode,&vgnd);
   if (colsread < 2)
      { 
       fprintf(stderr,"s2ibis: Apparent syntax error in input file. Pin type was GND.\n");
       exit(-1);
      }
   strncpyWarn(pinptr->nextlink->spice_node,spicenode, sizeof( pinptr->nextlink->spice_node ) );
   if (colsread == 3)
      {
       pinptr->nextlink->vil=vgnd;
      }
  }
else if (mod_type[0]=='N')
  {
   pinptr->nextlink->model_type=8;  
  }
else
  {
   fprintf(stderr,"s2ibis: (addpin)  unknown model type specified\n");
   fprintf(stderr,"s2ibis: offending input lines follow:\n");
   fprintf(stderr,"%s\n",ps1);
   fprintf(stderr,"%s\n",ps2);
   printpinlist(topofpinstack);
   exit(-1);
  }
return(pinptr->nextlink);
}


int addpinlist(FILE *fp, FILE *fq)
{
/* this routine reads through the input deck.  it finds the first and second lines */
/* for all the pins and calls addpin() to parse the lines and adds the data to the */
/* pin list.                                                                       */

/* this routine is called from the main program.  it expects fp to be pointing to  */
/* the first line of the [Pin] list. */
/* processing stops at the [Endpins] line. NC pins are assumed to have one-line    */
/* entries in the input file.   all others require two-line entries and may have   */
/* optional extra lines, which begin with the sequence **                          */


char finput[256];
char dummie[256];
char pin_string1[256];
char pin_string2[256];
PINPTR current_pin;
char *position;
int  colsread;

/* addpinlist assumes that fp is pointing to the first pin list line. */
/* from here on there can be no blank lines before the [Endpins] card */
/* just read in the lines, checking for POWER, GND, and NC cards, and */
/* pass on the contents to addpin.                                    */

if (fgets(finput,256,fp)==NULL)
   {
    fprintf(stderr,"s2ibis: (addpinlist) [Endpin] statement missing in input file\n");
    printpinlist(topofpinstack);
    exit(-1);
   }

while (strfind(finput,"[Endpin]") == NULL)
    {
     /* at this point we have a real live first pin line.                                  */
     /* strip off the asterisk, and store it in pin_string1.                               */     
     strncpyWarn(pin_string1,finput+1, sizeof( pin_string1) );

     /* now check to see if it has reserved word " NC " in it anywhere.  If it does, the   */
     /* next line to read in will be another first pin line                                */

     if (strfind(finput," NC "))
        {  /* yes, it is just a NC card.  print it, and just make up the second string     */
         fputs(pin_string1,fq);
         strncpyWarn(pin_string2,"NC", sizeof( pin_string2 ) );
         current_pin = addpin(pin_string1,pin_string2);
         if (fgets(finput,256,fp)==NULL)
            {
             fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Can't figure it out.\n");
             printpinlist(topofpinstack);
             exit(-1);
            }
        }
     else
        { /* no it wasn't an NC card. read another card and make sure it isn't a      */
          /* [Endpin] card.                                                           */
         if (fgets(finput,256,fp)==NULL)
            {
             fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Can't figure it out.\n");
             printpinlist(topofpinstack);
             exit(-1);
            }
         if (strfind(finput,"[Endpin]"))
            {
             fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Probably missing second line on an odd number of pins.\n");
             printpinlist(topofpinstack);
             exit(-1);
            }
         /* at this point we know it was a legitimate second line. create the pin list element*/
         strncpyWarn(pin_string2,finput+1, sizeof( pin_string2 ) );
         current_pin = addpin(pin_string1,pin_string2);

         /* this pin may have ** options.  get the next line and check for this. */
         if (fgets(finput,256,fp)==NULL)
            {
             fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Can't figure it out.\n");
             printpinlist(topofpinstack);
             exit(-1);
            }
         while ((finput[0] == 42) && (finput[1] == 42))  /* 42 is decimal for *   */
            {                                            /* so it starts with **  */
             if (position=strfind(finput,"C_comp"))
                {
                 if (strlen(position)>=8) strncpyWarn(current_pin->c_comp,position+6, sizeof( current_pin->c_comp ) );
                }
             else if (position=strfind(finput,"Vinh"))
                     {
                      if (strlen(position)>=6) strncpyWarn(current_pin->Vinh,position+4, sizeof( current_pin->Vinh ) );
                     }
             else if (position=strfind(finput,"Vinl"))
                     {
                      if (strlen(position)>=6) strncpyWarn(current_pin->Vinl,position+4, sizeof( current_pin->Vinl ) );
                     }
             else if (position=strfind(finput,"NoEnable"))
                     {
                      current_pin->show_enable=0;
                     }
             else if (position=strfind(finput,"Enable"))
                     {
                      current_pin->show_enable=2;
                      strncpyWarn(current_pin->enabletext,position, sizeof( current_pin->enabletext ));
                     }
             else if (position=strfind(finput,"NoPolarity"))
                     {
                      current_pin->show_polarity=0;
                     }
             else if (position=strfind(finput,"Polarity"))
                     {
                      current_pin->show_polarity=2;
                      strncpyWarn(current_pin->polartext,position, sizeof( current_pin->polartext ) );
                     }
             else if (position=strfind(finput,"NoModel"))
                     {
                      current_pin->no_model=1;
                     }
             else if (position=strfind(finput,"Input_ramp_typ"))
                     {
                      colsread=sscanf(position,"%s %f %f %le %le",dummie,&current_pin->vilrt,&current_pin->vihrt,&current_pin->trt,&current_pin->tft);
                      if (colsread != 5) 
                        {
                         fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Problem line follows:\n");
                         fprintf(stderr,"%s\n",finput);
                         printpinlist(topofpinstack);
                         exit(-1);
                        }
                     }
             else if (position=strfind(finput,"Input_ramp_min"))
                     {
                      colsread=sscanf(position,"%s %f %f %le %le",dummie,&current_pin->vilrn,&current_pin->vihrn,&current_pin->trn,&current_pin->tfn);
                      if (colsread != 5)
                        {
                         fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Problem line follows:\n");
                         fprintf(stderr,"%s\n",finput);
                         printpinlist(topofpinstack);
                         exit(-1);
                        }
                     }
             else if (position=strfind(finput,"Input_ramp_max"))
                     {
                      colsread=sscanf(position,"%s %f %f %le %le",dummie,&current_pin->vilrx,&current_pin->vihrx,&current_pin->trx,&current_pin->tfx);
                      if (colsread != 5)
                        {
                         fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Problem line follows:\n");
                         fprintf(stderr,"%s\n",finput);
                         printpinlist(topofpinstack);
                         exit(-1);
                        }
                     }
             else if (position=strfind(finput,"PrintPinList"))
                     {
                      printpinlist(topofpinstack);
                     }
             if (fgets(finput,256,fp)==NULL)
                {
                 fprintf(stderr,"s2ibis: (addpinlist) Error in pin list. Can't figure it out.\n");
                 printpinlist(topofpinstack);
                 exit(-1);
                }
            }
         if (!current_pin->no_model) fputs(pin_string1,fq);
        }
    }
return(1);
}

/* getpin returns a pointer to the pin entry corresponding to the pin name  */
/* given as an argument.                                                    */

PINPTR getpin(char targetpin[])
{
PINPTR pinptr;
pinptr=topofpinstack;
do
    {
     if (!strcmp(pinptr->pin,targetpin)) return(pinptr);
     pinptr=pinptr->nextlink;
    } while (pinptr);
fprintf(stderr,"s2ibis: (getpin) Caution!  Requested pin was not found!\n");
fprintf(stderr,"s2ibis: (getpin) Probably an error in the input deck.  \n");
exit(-1);
}

/* findground returns a pointer to the first GND type pin */
PINPTR findground()
{
PINPTR pinptr;
pinptr=topofpinstack;
do
    {
     if (pinptr->model_type == 7)
       return(pinptr);
     pinptr=pinptr->nextlink;
    } while (pinptr);
fprintf(stderr,"s2ibis: (findground()) Caution!  No GND pins in pin list!\n");
fprintf(stderr,"s2ibis: (findground()) Probably an error in the input deck.  \n");
exit(-1);
}

/* findpower returns a pointer to the first POWER pin */
PINPTR findpower()
{
PINPTR pinptr;
pinptr=topofpinstack;
do
    {
     if (pinptr->model_type == 6)
       return(pinptr);
     pinptr=pinptr->nextlink;
    } while (pinptr);
fprintf(stderr,"s2ibis: (findpower()) Caution!  No POWER pins in pin list!\n");
fprintf(stderr,"s2ibis: (findpower()) Probably an error in the input deck.  \n");
exit(-1);
}


int has_inpin(PINPTR targetpin)
{
 if (strlen(targetpin->inpin)>0) return(1);
 else return(0);
}

int has_enable(PINPTR targetpin)
{
 if (strlen(targetpin->enable)>0) return(1);
 else return(0);
}

/* addmod is used (by makemodel) to avoid running simulations for pins that have the same */
/* model_type as previously encountered pins.                                             */

int addmod(char newmod[])
{

MODPTR modptr;

/* see if this mod is on stack. if so return 1, else add to stack and return 0 */

modptr=topofmodstack;
while (modptr->nextlink) {
			  if (!strcmp(modptr->nextlink->model_name,newmod)) return(1);
                          else modptr=modptr->nextlink;
			 }
modptr->nextlink=(struct modentry *) calloc(1,sizeof(MODELEM));
modptr->nextlink->lastlink=modptr;
strncpyWarn(modptr->nextlink->model_name,newmod, sizeof( modptr->nextlink->model_name ));
if (modptr->nextlink) return(0);
else
 {
  fprintf(stderr,"s2ibis: Unable to allocate memory for new model entry. Exiting. \n");
  exit(-1);
 }
}
