/*S2IBIS Version 1.0 - Steven Lipa, NC State University  August 4, 1994       */

/******************************************************************************/
/*  Copyright 1994, North Carolina State University.                          */
/*                                                                            */
/*  North Carolina State University (NCSU) and the author hereby disclaim     */
/*  all implied warranties, including the implied warranties of merchant-     */
/*  ability and fitness for a particular purpose. NCSU is not liable for      */
/*  any damages incurred by the use of this software and/or documentation,    */
/*  including direct, indirect, special, incidental, or consequential         */
/*  damages.                                                                  */
/*                                                                            */
/*  NCSU grants the bearer the right to modify, copy, and redistribute        */
/*  this software and documentation, but requests that NCSU be acknowledged   */
/*  for its contribution in the resulting product.                            */
/******************************************************************************/

/******************************* RELEASE NOTES ********************************/

/* 3/22/94 initial release Version 0.0 */

/* 3/27/94  Version 0.1 */
/* added Ramp tables */
/* added non-integer SPICE NODE compatibility */
/* added some debugging aids to the pins.c routines */

/* 5/31/94   Version 0.2 */
/* variable # of steps allowed for DC tables */
/* non-integer pin names added */
/* bug in pull-up and power-clamp tables fixed */
/* pull-up/power-clamp partitioning added */
/* bug causing re-simulation of already completed models fixed */
/* vil-vih-input-ramp-speed parameters added for Ramp simulations */
/* HSPICE capabilities added (but not fully tested!) */

/* 7/1/94 Version 0.3*/
/* Bug fixes to partitioning implemented */
/* HSPICE compatibility augmented */

/* 7/14/94 Version 0.4 */
/* alphanumeric pin names allowed */
/* changed from header line to header section in input file */
/* capability to add optional [Notes] [Disclaimer] [Source] to IBIS header added */
/* capability to add optional [File Rev] to IBIS header added                    */
/* capability to add optional parameters (Vinh, Vinl, C_comp) to [Model]s added  */
/* capability to turn off Polarity and Enable parameters in [Model]s added       */
/* changed method for specifying spice type                                      */

/* 7/25/94 Version 0.5 */
/* further refinement of partitioning scheme  */
/* added [Clamp_Tolerance]    */
/* added [Iterate] capability    */
/* added multiple input model capability    */
/* added code to omit trivial clamp tables    */
/* added automatic pulse amplitude adjustment for min and max sims    */
/* added manual override pulse amplitude adjustment for min and max sims    */
/* added **Polarity **NoPolarity **Enable **NoEnable keywords to header    */
/* added some cosmetic changes to output files    */
/* added multiple input model capability    */
/* added the **NoModel option */

/* 7/26/94 Version 0.6 */
/* added the [Package] code  */
/* added check for Output-only models to always omit clamp tables  */
/* fixed bug in [Numpoints], [Clamp_Tolerance] */
/* changed clamp table spacings */
/* changed cosmetics around [Package] and [Pin] */

/* 7/27/94 Version 0.7 */
/* fixed/enhanced aborted SPICE run detection and reporting */
/* fixed bug in power clamp table generation (was supressing based on ground clamp results...) */
/* added headers in all s2ibis-generated SPICE decks */
/* more cosmetic changes */

/* 7/30/94 Version 0.7a */
/* added pinclean() function to deallocate pin stack upon completion */
/* eliminated SPICE jobs for Output-only clamps.  was just ignoring output before */
/* fixed bug in aborted tables.  IBIS reserved word NA substituted for N/A        */ 
/* added warning when ramp output transition < supply/2.0 (V)  */
/* more cosmetic fixes */
/* fixed bug in multiple-input implementation */ 

/* 8/2/94 Version 0.7b */
/* changed calling sequence for HSPICE and SPICE3 to allow for message file */
/* added check for aborted clamp sim in generating I/O and 3-state pullup and pulldown tables */
/* added strfind function for better portability.  strfind is my version of strstr */
/* made the [Disclaimer], [Notes], etc. use dynamically allocated temporary space and removed limits */
/* fixed another bug in NOTKEYA NOTKEYB.  added File Rev and Iterate keywords, removed Numpoints */
/* removed all references to numpoints, as they are now obsolete */
 
/* 8/3/94 Version 0.7c */
/* added modclean() function to fix memory leak*/
/* changed call to PSPICE to suppress status window */
/* changed getrampdata and getspicedata to improve ability to detect aborted PSPICE simulations */

/* 8/4/94 Version 1.0 */
/* Initial public release */

#include "s2ibis.h"

#define NOTKEYA ( !strfind(finputa,"[Pin]") && !strfind(finputa,"[Spice]") && !strfind(finputa,"[Source]") && !strfind(finputa,"[Disclaimer]") && !strfind(finputa,"[Notes]" ))

#define NOTKEYB (!strfind(finputa,"[Package]") && !strfind(finputa,"[Clamp_Tolerance]") && !strfind(finputa,"[Iterate]")     && !strfind(finputa,"[File Rev]"))



MODPTR topofmodstack;                /* global variable pointing to top of model list*/
PINPTR topofpinstack;                /* global variable pointing to top of pin stack */
NODEPTR topofnodestack;              /* global variable pointing to top of node stack */
int spicetype;                       /* global variable for designating SPICE version*/
                                     /* 2=Berk 2g6 */
                                     /* 3=Berk 3f4 */
                                     /* 4=PSPICE   */
                                     /* 5=HSPICE   */
int iterate;                         /* used to suppress SPICE run if output file exists */
double clmptol;                      /* variable clamp tolerance limit used to suppress clamp tables */

void textpush(char *thetext, TEXTPTR tptr)
{

while (tptr->nextlink) tptr=tptr->nextlink;
tptr->nextlink = (struct textentry *) calloc(1,sizeof(TEXTELEM));
if (tptr->nextlink == NULL)
   {
    fprintf(stderr,"s2ibis: Unable to allocate memory.  Exiting.\n");
    exit(-1);
   }
tptr->nextlink->lastlink=tptr;
tptr->nextlink->nextlink=NULL;
strncpy(tptr->nextlink->textline,thetext,81);
}

void textclean(TEXTPTR tptr)
{
TEXTPTR temptr;

temptr=tptr;
tptr=tptr->nextlink;
temptr->nextlink=NULL;
while (tptr)
    {
     temptr=tptr;
     tptr=tptr->nextlink;
     free(temptr);
    }
}


void truncate(char trunk[], int tlen)  /* truncates strings > tlen chars */
{
 if (strlen(trunk)>tlen)
    {
     *(trunk+tlen-1)='\n';
     *(trunk+tlen)='\0';
    }
}

char *strfind (char *instr, char *targstr)
{
char *sa, *sb;
int sidx;

if (strlen(targstr) > strlen(instr)) return(NULL);
while (*(instr+strlen(targstr)-1))
     {
      sa=instr;
      sb=targstr;
      while ((*sa == *sb)&&(sa < instr+strlen(targstr)))
         {
          sa++;
          sb++;
         }
      if (sa == instr+strlen(targstr)) return(instr);
      instr++;
     }
return(NULL);
}

void strwrem(char *s) /* removes leading whitespace from a string in place */
{
 char *t;

 t=s; 
 while( (*t == ' ') || (*t == '\t')) t++;
 while(*s++ = *t++) /* EMPTY */ ;
}

int addheader(FILE *fp, FILE *fq,char outfilename[])
{
/* this routine starts building the ibis model file. The
   following parts are added:
   [IBIS Ver]
   [Comment char]
   [File name]
   [File Rev]
   [Date]
   [Source]
   [Disclaimer]
   [Component]
   [Manufacturer]
   [Package]     
   
   The routine returns the value 1 if the model is CMOS and 2 if BIPOLAR
   A negative value is returned if a syntax error is detected in the input file,
   but the error checking is primitive.                                                         */ 

char device_name[42]; 
char company_name[42];
char package_type[82];
char tek[256];
char finputa[256];           /* temporary storage area for reading in strings from input file */
char tstring[256];           /* temporary storage area for making up strings */
time_t timeofday;
int filecheck;               /* used to check for optional entries in header */
int discheck;                /* used to check for optional entries in header */
int pakcheck;                /* used to check for optional entries in header */
int notescheck;              /* used to check for optional entries in header */
int sourcecheck;             /* used to check for optional entries in header */
char file_rev[256];
char *position;
TEXTELEM disctext,notetext,packtext,sourtext; /* dynamic storage for user text */
TEXTPTR textptr;                              /* dummy variable to point to various TEXTELEMs */
int needmore;
int idx;
int namelength; /* used for centering header comment */
int devlength; /* used for centering header comment */

filecheck=discheck=sourcecheck=notescheck=pakcheck=0;
spicetype=2; /* Berkeley SPICE 2 is the default version.  To override use *[Spice] spicetype */
iterate=0;  /* iterate switch defaults to zero unless user uses *[Iterate] */

/* initialize dynamically-allocated text storage areas */
disctext.lastlink = disctext.nextlink = NULL;
notetext.lastlink = notetext.nextlink = NULL;
sourtext.lastlink = sourtext.nextlink = NULL;
packtext.lastlink = packtext.nextlink = NULL;

/* read first line of input file to determine company_name */
fgets(finputa,256,fp);
if (strlen(finputa+1) > 41)
   {
    fprintf(stderr,"s2ibis: Manufacturer's name cannot exceed 40 characters.\n");
    fprintf(stderr,"s2ibis: Truncating to 40 characters! Sorry, see the IBIS\n");
    fprintf(stderr,"s2ibis: Version 1.1 spec for details. (Not my idea....) \n");
    truncate(finputa+1,41);
   }
strncpyWarn(company_name,finputa+1, sizeof( company_name ) );

/* read next line of input file to determine device_name   */
fgets(finputa,256,fp);
if (strlen(finputa+1) > 41)
   {
    fprintf(stderr,"s2ibis: Device name cannot exceed 40 characters.\n");
    fprintf(stderr,"s2ibis: Truncating to 40 characters! Sorry, see the IBIS\n");
    fprintf(stderr,"s2ibis: Version 1.1 spec for details. (Not my idea....) \n");
    truncate(finputa+1,41);
   }
strncpyWarn(device_name,finputa+1, sizeof( device_name ) );

/* read next line of input file to determine package_type  */
fgets(finputa,256,fp);
strncpyWarn(package_type,finputa+1, sizeof( package_type ) );

/* read next line of input file to determine technology    */
fgets(finputa,256,fp);
strncpyWarn(tek,finputa+1, sizeof( tek ) );

/* check for the optional keywords */

fgets(finputa,256,fp);
while (!strfind(finputa,"[Pin]"))
      {
       needmore = 1;
       if (strfind(finputa,"[Spice]"))
          {
           needmore=1;
           if (strfind(finputa,"2")) spicetype=2;
           else if (strfind(finputa,"3")) spicetype=3;
           else if (strfind(finputa,"P")) spicetype=4;
           else if (strfind(finputa,"H")) spicetype=5;
           else
               {
                fprintf(stderr,"s2ibis (addheader) Illegal SPICE version specified.\n");
                return(-1);
               }
          }
       else if (position=strfind(finputa,"[File Rev]"))
               {
                strwrem(position+10);
                needmore=1;
                filecheck=1;
                if (strlen(position+10) > 69) truncate(position+10,69);
                strncpyWarn(file_rev,position+10, sizeof( file_rev ) );
               }
       else if (position=strfind(finputa,"[Iterate]"))
               {
                needmore=1;
                iterate=1;
               }
       else if (position=strfind(finputa,"[Clamp_Tolerance]"))
               {
                needmore=1;
                sscanf(position+17,"%le",&clmptol);
               }
       else if (position=strfind(finputa,"[Source]"))
               {
                sourcecheck=1;
                if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                for(idx=0;idx<=7;idx++) *position++=' ';
                textpush(finputa+1,&sourtext);
                fgets(finputa,256,fp);
                while (NOTKEYA && NOTKEYB)
                   {
                    if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                    textpush(finputa+1,&sourtext);
                    fgets(finputa,256,fp);
                   } 
                needmore=0;
               }
       else if (strfind(finputa,"[Disclaimer]"))
               {
                discheck=1;
                if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                textpush(finputa+1,&disctext);
                fgets(finputa,256,fp);
                while (NOTKEYA && NOTKEYB)
                   {
                    if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                    textpush(finputa+1,&disctext);
                    fgets(finputa,256,fp);
                   }
                needmore=0;
               }
       else if (strfind(finputa,"[Notes]"))
               {
                notescheck=1;
                if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                textpush(finputa+1,&notetext);
                fgets(finputa,256,fp);
                while (NOTKEYA && NOTKEYB)
                   {
                    if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                    textpush(finputa+1,&notetext);
                    fgets(finputa,256,fp);
                   }
                needmore=0;
               }
       else if (strfind(finputa,"[Package]"))
               {
                pakcheck=1;
                if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                textpush(finputa+1,&packtext);
                fgets(finputa,256,fp);
                while (NOTKEYA && NOTKEYB)
                   {
                    if (strlen(finputa+1) > 81) truncate(finputa+1,81);
                    textpush(finputa+1,&packtext);
                    fgets(finputa,256,fp);
                   }
                needmore=0;
               }
       if (needmore) fgets(finputa,256,fp);
      }
namelength=strlen(outfilename);
devlength=strlen(device_name);
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|");
for (idx=0;idx<=(38-(namelength+devlength+24)/2);idx++) fprintf(fq," ");
fprintf(fq,"* %s * IBIS 1.1 Model for %s",outfilename,device_name);
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|\n");
fprintf(fq,"[IBIS Ver]\t1.1\n");
fprintf(fq,"[Comment char]\t|_char\n");
fprintf(fq,"[File name]\t%s\n",outfilename);
timeofday=time(NULL);
fprintf(fq,"[Date]\t\t%s",ctime(&timeofday));

if (!filecheck)
     {
      fprintf(fq,"[File Rev]\t0.0\n");
     }
else 
     {
      strncpyWarn(tstring,"[File Rev]\t",sizeof( tstring ) );
      strncat(tstring,file_rev,50);
      fprintf(fq,"%s",tstring);
     }
fprintf(fq,"[Source] s2ibis Version 1.0 SPICE to IBIS converter\n");
if (sourcecheck)
     {
      textptr=sourtext.nextlink;
      while (textptr)
         {
          fprintf(fq,"%s",textptr->textline);
          textptr=textptr->nextlink;
         }
      textclean(&sourtext);
     }
if (notescheck)
     {
      textptr=notetext.nextlink;
      while (textptr)
         {
          fprintf(fq,"%s",textptr->textline);
          textptr=textptr->nextlink;
         }
      textclean(&notetext);
     }
if (discheck)
     {
      textptr=disctext.nextlink;
      while (textptr)
         {
          fprintf(fq,"%s",textptr->textline);
          textptr=textptr->nextlink;
         }
      textclean(&disctext);
     }

switch (spicetype)
    {
     case 3: fprintf(stderr,"s2ibis: Specified simulator: Berkeley SPICE 3.\n");
             break;
     case 4: fprintf(stderr,"s2ibis: Specified simulator: PSPICE.\n");
             break;
     case 5: fprintf(stderr,"s2ibis: Specified simulator: HSPICE.\n");
             break;
     default: fprintf(stderr,"s2ibis: Default simulator: Berkeley SPICE 2.\n");
              break;
    }

fprintf(fq,"|\n");
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|");
for (idx=0;idx<=(38-(devlength+11)/2);idx++) fprintf(fq," ");
fprintf(fq,"COMPONENT: %s",device_name);
fprintf(fq,"|*******************************************************************************\n");
fprintf(fq,"|\n");

strncpyWarn(tstring,"[Component]\t", sizeof( tstring ) );
strncat(tstring,device_name,41);
fprintf(fq,"%s",tstring);

strncpyWarn(tstring,"[Manufacturer]\t", sizeof( tstring ) );
strncat(tstring,company_name,41);
fprintf(fq,"%s",tstring);


if (pakcheck)
   {
    textptr=packtext.nextlink;
    while (textptr)
       {
        fprintf(fq,"%s",textptr->textline);
        textptr=textptr->nextlink;
       }
    textclean(&packtext);
   }
else
   {
    fprintf(fq,"[Package]\n");
    fprintf(fq,"| These are default values which must be edited by the user for now....\n");
    fprintf(fq,"| variable            typ              min              max\n");
    fprintf(fq,"R_pkg                 0.0m             0.0m             0.0m\n");
    fprintf(fq,"L_pkg                 0.0nH            0.0nH            0.0nH\n");
    fprintf(fq,"C_pkg                 0.0pF            0.0pF            0.0pF\n");
   }

fputs("|\n",fq);
fputs("|*******************************************************************************\n",fq);
fputs("|\n",fq);

/* fp should be pointing to the [Pin] line.  Add this line to the output deck to */
/* prepare for calling the routine that will generate the pin table              */

fputs(finputa+1,fq);

if ((tek[0]=='C') || (tek[0]=='c')) return(1);
else if ((tek[0]=='B') || (tek[0]=='b')) return(2);
        else return(-2);
}

void pinclean()
{
PINPTR nptr;
PINPTR tptr;

nptr=topofpinstack->nextlink;
topofpinstack->nextlink=NULL;
while (nptr)
    {
     tptr=nptr;
     nptr=nptr->nextlink;
     free(tptr);
    }
}

void modclean()
{
MODPTR nptr;
MODPTR tptr;

nptr=topofmodstack->nextlink;
topofmodstack->nextlink=NULL;
while (nptr)
    {
     tptr=nptr;
     nptr=nptr->nextlink;
     free(tptr);
    }
}


main (int argc,char *argv[])
{

FILE *fopen();                       /* routine used to open files */
FILE *fp;                            /* input filename file pointer */
FILE *fq;                            /* output filename file pointer */
FILE *fd;                            /* input spice deck temporary storage */
char outfilename[256];               /* temp storage for making output filename */
PINELEM pintop;                      /* dummy element for top of pin stack */
MODELEM modtop;                      /* dummy element for top of mod list */
NODEELEM nodetop;                    /* dummy element for top of node list */
int  technology;                     /* 1=CMOS 2=BIPOLAR */
char finput[256];                     /* input string used in copying SPICE deck to fd */
char *strtemp;                       /* used to do pointer math checking for .END cards */
char input_file[256];


clmptol=1.0E-6; /* set up the default minimum clamp current */
/* check for correct syntax */
if ((argc < 2) || (argc > 3))
  {
   fprintf(stderr,"Usage:   s2ibis input_filename [output_filename]\n");
   exit(-1);
  }

/* this is the beginning of the executable section */
fprintf(stderr,"s2ibis: Version 1.0 North Carolina State University\n");

/* open input file just to make sure it is there!  report, and abort if unable */
if ((fp = fopen(*++argv,"r")) == NULL)
  {
   fprintf(stderr,"s2ibis: unable to open %s\n",*argv);
   exit(-1);
  }
fprintf(stderr,"s2ibis: Using input file named %s\n",*argv);
strncpyWarn(input_file,*argv, sizeof( input_file ));
argc--;

/* open outputfile, if specified. */

if (argc < 2)
  {
   /* no output file was specified.  output goes to stdout */
   fq=stdout;
   fprintf(stderr,"s2ibis: Output going to stdout.\n");
  }
else  /* an outputfilename was provided on the command line. open. */
  {
   if ((fq = fopen(*++argv,"w")) == NULL) 
      {
       fprintf(stderr,"s2ibis: Unable to open %s.\n",*argv);
      }
   else
      {
       fprintf(stderr,"s2ibis: Using output file named %s\n",*argv);
       strncpyWarn(outfilename,*argv, sizeof( outfilename ) );
      }
  }

/*** At this point, fp points to the input deck and fq points to the xxx.ibs output file ***/

/* Add the ibis header */
technology=addheader(fp,fq,outfilename);
if (technology<0) 
   {
    fprintf(stderr,"s2ibis: Error: technology must be CMOS or BIPOLAR. Exiting.\n");
    exit(-1);
   }

 
/* initialize mod list structure */
topofmodstack = &modtop;
topofmodstack->lastlink = NULL;
topofmodstack->nextlink = NULL;

/* initialize node list structure */
topofnodestack = &nodetop;
topofnodestack->lastlink = NULL;
topofnodestack->nextlink = NULL;

/* initialize pin list structure */
topofpinstack = &pintop;
topofpinstack->lastlink = NULL;
topofpinstack->nextlink = NULL;
addpinlist(fp,fq);
fprintf(fq,"|\n");
/* printpinlist(topofpinstack); */

/* at this point: */
/*        the pin list has been created                 */
/*        the .ibs file has its header and [Pin] sections in it */
/*        the fp file pointer is pointing to the beginning of the actual SPICE deck */

/* reopen input file */
/* copy the remaining SPICE deck to temporary file s2ibis.tmp using file pointer fd*/
/* do NOT copy the .END card. */
if ((fd = fopen("s2ibis.tmp","w")) == NULL)
  {
   fprintf(stderr,"s2ibis: unable to open temporary storage file s2ibis.tmp. aborting.\n");
   exit(-1);
  }

while (fgets(finput,256,fp) != NULL)
      {
       strtemp=strfind(finput,".END");
       if (strtemp)
          {
           /* then the string .END is on this card. make sure it is not a .ENDS card, */
           /* and if it's not, stop copying.  also, check to see if it's a comment    */
           /* card that just mentions the .END card!                                  */
           if ( (*finput!= '*' ) &&  (*(strtemp+4)!='S') && (*(strtemp+4)!='s')  )
             break; 
          }
       fputs(finput,fd);
      }
/* the input file is no longer needed.  close it */
fclose(fp);
fclose(fd);

/* add the models to the .ibs file */
addmodels(fq,technology);

fputs("[End]\n",fq);
fclose(fq);

pinclean();
modclean();
}
